@@ -106,7 +106,6 @@ npth_clock_gettime(struct timespec *tp)
 {
   FILETIME ftime;
   ULARGE_INTEGER systime;
-  unsigned long long usecs;
 
   GetSystemTimeAsFileTime (&ftime);
   systime.LowPart = ftime.dwLowDateTime;
@@ -150,8 +149,6 @@ calculate_timeout (const struct timespec *abstime, DWORD *msecs_r)
 static void
 enter_npth (const char *function)
 {
-  int res;
-
   if (DEBUG_CALLS)
     _npth_debug (DEBUG_CALLS, "tid %lu: enter_npth (%s)\n",
 		 npth_self (), function ? function : "unknown");
@@ -175,8 +172,19 @@ leave_npth (const char *function)
 #define LEAVE() leave_npth(__FUNCTION__)
 
 
+struct dl {
+  struct dl *next;
+  struct dl *prev;
+};
+
 struct npth_impl_s
 {
+  /* Doubly-linked list for the waiter queue in condition
+     variables.  */
+  /* Should be the first member of struct npth_impl_s, so that
+   * casting from the address of l can be the head of struct npth_impl_s.  */
+  struct dl l;
+
   /* Usually there is one ref owned by the thread as long as it is
      running, and one ref for everybody else as long as the thread is
      joinable.  */
@@ -193,11 +201,6 @@ struct npth_impl_s
 
   char name[THREAD_NAME_MAX + 1];
 
-  /* Doubly-linked list for the waiter queue in condition
-     variables.  */
-  npth_impl_t next;
-  npth_impl_t *prev_ptr;
-
   /* The event on which this thread waits when it is queued.  */
   HANDLE event;
 
@@ -206,32 +209,35 @@ struct npth_impl_s
 
 
 static void
-dequeue_thread (npth_impl_t thread)
+dequeue_thread (struct dl *l)
 {
-  /* Unlink the thread from any condition waiter queue.  */
-  if (thread->next)
+  /* Unlink the entry of L from a condition waiter queue.  */
+  struct dl *next = l->next;
+  struct dl *prev = l->prev;
+
+  if (next)
     {
-      thread->next->prev_ptr = thread->prev_ptr;
-      thread->next = NULL;
+      next->prev = prev;
+      l->next = NULL;
     }
-  if (thread->prev_ptr)
+
+  if (prev)
     {
-      *(thread->prev_ptr) = thread->next;
-      thread->prev_ptr = NULL;
+      prev->next = next;
+      l->prev = NULL;
     }
 }
 
 
-/* Enqueue THREAD to come after the thread whose next pointer is
-   prev_ptr.  */
 static void
-enqueue_thread (npth_impl_t thread, npth_impl_t *prev_ptr)
+enqueue_thread (struct dl *l, struct dl *head)
 {
-  if (*prev_ptr)
-    (*prev_ptr)->prev_ptr = &thread->next;
-  thread->prev_ptr = prev_ptr;
-  thread->next = *prev_ptr;
-  *prev_ptr = thread;
+  struct dl *last_entry = head->prev;
+
+  head->prev = l;
+  l->prev = last_entry;
+  l->next = head;
+  last_entry->next = l;
 }
 
 
@@ -267,15 +273,15 @@ new_thread (npth_t *thread_id)
     return errno;
 
   thread->refs = 1;
-  thread->handle = INVALID_HANDLE_VALUE;
+  thread->handle = NULL;
   thread->detached = 0;
   thread->start_routine = NULL;
   thread->start_arg = NULL;
-  thread->next = NULL;
-  thread->prev_ptr = NULL;
+  thread->l.next = NULL;
+  thread->l.prev = NULL;
   /* We create the event when it is first needed (not all threads wait
      on conditions).  */
-  thread->event = INVALID_HANDLE_VALUE;
+  thread->event = NULL;
   memset (thread->name, '\0', sizeof (thread->name));
 
   thread_table[id] = thread;
@@ -293,8 +299,11 @@ free_thread (npth_t thread_id)
   if (thread->handle)
     CloseHandle (thread->handle);
 
+  if (thread->event)
+    CloseHandle (thread->event);
+
   /* Unlink the thread from any condition waiter queue.  */
-  dequeue_thread (thread);
+  dequeue_thread (&thread->l);
 
   free (thread);
 
@@ -351,7 +360,7 @@ npth_init (void)
   thread = thread_table[thread_id];
   thread->handle = handle;
 
-  if (! TlsSetValue(tls_index, (LPVOID) thread_id))
+  if (! TlsSetValue(tls_index, (LPVOID)(uintptr_t) thread_id))
     return map_error (GetLastError());
 
   LEAVE();
@@ -384,6 +393,7 @@ int
 npth_attr_destroy (npth_attr_t *attr)
 {
   free (*attr);
+  *attr = NULL;
   return 0;
 }
 
@@ -448,11 +458,11 @@ npth_setname_np (npth_t target_thread, const char *name)
 static DWORD
 thread_start (void *arg)
 {
-  npth_t thread_id = (npth_t) arg;
+  npth_t thread_id = (npth_t)(uintptr_t) arg;
   npth_impl_t thread;
   void *result;
 
-  if (! TlsSetValue(tls_index, (LPVOID) thread_id))
+  if (! TlsSetValue(tls_index, (LPVOID)(uintptr_t) thread_id))
     /* FIXME: There is not much we can do here.  */
     ;
 
@@ -482,33 +492,18 @@ npth_create (npth_t *newthread, const npth_attr_t *user_attr,
 {
   int err = 0;
   npth_t thread_id = INVALID_THREAD_ID;
-  npth_attr_t attr;
-  int attr_allocated;
   npth_impl_t thread;
   HANDLE handle;
 
   /* We must stay protected here, because we access the global
      thread_table.  Also, creating a new thread is not a blocking
      operation.  */
-  if (user_attr)
-    {
-      attr = *user_attr;
-      attr_allocated = 0;
-    }
-  else
-    {
-      err = npth_attr_init (&attr);
-      if (err)
-	return err;
-      attr_allocated = 1;
-    }
-
   err = new_thread (&thread_id);
   if (err)
     goto err_out;
 
   thread = thread_table[thread_id];
-  if (attr->detachstate == NPTH_CREATE_DETACHED)
+  if (user_attr && (*user_attr)->detachstate == NPTH_CREATE_DETACHED)
     thread->detached = 1;
   else
     thread->refs += 1;
@@ -518,7 +513,7 @@ npth_create (npth_t *newthread, const npth_attr_t *user_attr,
 
   handle = CreateThread (NULL, 0,
 			 (LPTHREAD_START_ROUTINE)thread_start,
-			 (void *) thread_id, CREATE_SUSPENDED,
+			 (void *)(uintptr_t) thread_id, CREATE_SUSPENDED,
 			 NULL);
   if (handle == NULL)
     {
@@ -536,8 +531,6 @@ npth_create (npth_t *newthread, const npth_attr_t *user_attr,
  err_out:
   if (thread_id)
     free_thread (thread_id);
-  if (attr_allocated)
-    npth_attr_destroy (&attr);
 
   return err;
 }
@@ -552,7 +545,7 @@ npth_self (void)
   if (thread_id == 0 && GetLastError() != ERROR_SUCCESS)
     /* FIXME: Log the error.  */
     ;
-  return (npth_t) thread_id;
+  return (npth_t)(uintptr_t) thread_id;
 }
 
 
@@ -562,7 +555,6 @@ npth_tryjoin_np (npth_t thread_id, void **thread_return)
 {
   int err;
   npth_impl_t thread;
-  int res;
 
   err = find_thread (thread_id, &thread);
   if (err)
@@ -593,7 +585,6 @@ npth_join (npth_t thread_id, void **thread_return)
 {
   int err;
   npth_impl_t thread;
-  int res;
 
   /* No need to allow competing threads to enter when we can get the
      lock immediately.  */
@@ -717,8 +708,19 @@ int
 npth_setspecific (npth_key_t key, const void *pointer)
 {
   BOOL res;
+  union {
+    const void *p_readonly;
+    void *p_writable;
+  } p;
 
-  res = TlsSetValue (key, (void *) pointer);
+  /*
+   * In the API of TlsSetValue, the second argument type is a pointer
+   * (writable object).  It were better to have a pointer type
+   * (readonly object), as in the POSIX thread API.  To express that
+   * it's our intentional type casting, we use union here.
+   */
+  p.p_readonly = pointer;
+  res = TlsSetValue (key, p.p_writable);
   if (res == 0)
     return map_error (GetLastError());
 
@@ -834,7 +836,7 @@ static int
 mutex_init_check (npth_mutex_t *mutex)
 {
   int err;
-  npth_mutexattr_t attr;
+  npth_mutexattr_t attr = NULL;
   int kind;
 
   if (*mutex == 0)
@@ -902,7 +904,6 @@ int
 npth_mutex_trylock (npth_mutex_t *mutex)
 {
   int err;
-  DWORD res;
 
   /* While we are protected, let's check for a static initializer.  */
   err = mutex_init_check (mutex);
@@ -976,7 +977,7 @@ struct npth_cond_s
      simple.  */
 
   /* The waiter queue.  */
-  npth_impl_t waiter;
+  struct dl waiter;
 };
 
 
@@ -994,7 +995,7 @@ npth_cond_init (npth_cond_t *cond_r,
   if (!cond)
     return errno;
 
-  cond->waiter = NULL;
+  cond->waiter.next = cond->waiter.prev = &cond->waiter;
 
   *cond_r = cond;
   return 0;
@@ -1007,7 +1008,7 @@ npth_cond_destroy (npth_cond_t *cond)
   if (*cond == 0)
     return EINVAL;
 
-  if ((*cond)->waiter)
+  if ((*cond)->waiter.next != &(*cond)->waiter)
     return EBUSY;
 
   free (*cond);
@@ -1045,18 +1046,20 @@ npth_cond_signal (npth_cond_t *cond)
   int err;
   npth_impl_t thread;
   DWORD res;
+  struct dl *waiter;
 
   /* While we are protected, let's check for a static initializer.  */
   err = cond_init_check (cond);
   if (err)
     return err;
 
-  if ((*cond)->waiter == INVALID_THREAD_ID)
+  waiter = (*cond)->waiter.next;
+  if (waiter == &(*cond)->waiter)
     return 0;
 
   /* Dequeue the first thread and wake it up.  */
-  thread = (*cond)->waiter;
-  dequeue_thread (thread);
+  dequeue_thread (waiter);
+  thread = (struct npth_impl_s *)waiter;
 
   res = SetEvent (thread->event);
   if (res == 0)
@@ -1083,26 +1086,26 @@ npth_cond_broadcast (npth_cond_t *cond)
   int err;
   npth_impl_t thread;
   DWORD res;
-  int any;
+  struct dl *waiter;
 
   /* While we are protected, let's check for a static initializer.  */
   err = cond_init_check (cond);
   if (err)
     return err;
 
-  if ((*cond)->waiter == INVALID_THREAD_ID)
-    return 0;
-
-  while ((*cond)->waiter)
+  waiter = (*cond)->waiter.next;
+  while (waiter != &(*cond)->waiter)
     {
       /* Dequeue the first thread and wake it up.  */
-      thread = (*cond)->waiter;
-      dequeue_thread (thread);
+      dequeue_thread (waiter);
+      thread = (struct npth_impl_s *)waiter;
 
       res = SetEvent (thread->event);
       if (res == 0)
 	/* FIXME: An error here implies a mistake in the npth code.  Log it.  */
 	;
+
+      waiter = (*cond)->waiter.next;
     }
 
   /* Force the woken up threads into the mutex lock function (for the
@@ -1132,7 +1135,6 @@ npth_cond_wait (npth_cond_t *cond, npth_mutex_t *mutex)
   int err2;
   BOOL bres;
   npth_impl_t thread;
-  npth_impl_t *prev_ptr;
 
   /* While we are protected, let's check for a static initializer.  */
   err = cond_init_check (cond);
@@ -1144,18 +1146,14 @@ npth_cond_wait (npth_cond_t *cond, npth_mutex_t *mutex)
     return err;
 
   /* Ensure there is an event.  */
-  if (thread->event == INVALID_HANDLE_VALUE)
+  if (thread->event == NULL)
     {
       thread->event = CreateEvent (NULL, TRUE, FALSE, NULL);
-      if (thread->event == INVALID_HANDLE_VALUE)
+      if (thread->event == NULL)
 	return map_error (GetLastError());
     }
 
-  /* Find end of queue and enqueue the thread.  */
-  prev_ptr = &(*cond)->waiter;
-  while (*prev_ptr)
-    prev_ptr = &(*prev_ptr)->next;
-  enqueue_thread (thread, prev_ptr);
+  enqueue_thread (&thread->l, &(*cond)->waiter);
 
   /* Make sure the event is not signaled before releasing the mutex.  */
   bres = ResetEvent (thread->event);
@@ -1168,7 +1166,7 @@ npth_cond_wait (npth_cond_t *cond, npth_mutex_t *mutex)
       err = npth_mutex_unlock (mutex);
       if (err)
 	{
-	  dequeue_thread (thread);
+	  dequeue_thread (&thread->l);
 	  return err;
 	}
     }
@@ -1178,7 +1176,7 @@ npth_cond_wait (npth_cond_t *cond, npth_mutex_t *mutex)
   LEAVE();
 
   /* Make sure the thread is dequeued (in case of error).  */
-  dequeue_thread (thread);
+  dequeue_thread (&thread->l);
 
   if (mutex)
     {
@@ -1203,7 +1201,6 @@ npth_cond_timedwait (npth_cond_t *cond, npth_mutex_t *mutex,
   int err2;
   BOOL bres;
   npth_impl_t thread;
-  npth_impl_t *prev_ptr;
   DWORD msecs;
 
   err = calculate_timeout (abstime, &msecs);
@@ -1236,10 +1233,10 @@ npth_cond_timedwait (npth_cond_t *cond, npth_mutex_t *mutex,
     return err;
 
   /* Ensure there is an event.  */
-  if (thread->event == INVALID_HANDLE_VALUE)
+  if (thread->event == NULL)
     {
       thread->event = CreateEvent (NULL, TRUE, FALSE, NULL);
-      if (thread->event == INVALID_HANDLE_VALUE)
+      if (thread->event == NULL)
 	return map_error (GetLastError());
     }
 
@@ -1249,16 +1246,12 @@ npth_cond_timedwait (npth_cond_t *cond, npth_mutex_t *mutex,
     /* Log an error.  */
     ;
 
-  /* Find end of queue and enqueue the thread.  */
-  prev_ptr = &(*cond)->waiter;
-  while (*prev_ptr)
-    prev_ptr = &(*prev_ptr)->next;
-  enqueue_thread (thread, prev_ptr);
+  enqueue_thread (&thread->l, &(*cond)->waiter);
 
   err = npth_mutex_unlock (mutex);
   if (err)
     {
-      dequeue_thread (thread);
+      dequeue_thread (&thread->l);
       return err;
     }
 
@@ -1369,20 +1362,6 @@ npth_rwlock_init (npth_rwlock_t *rwlock_r,
 {
   int err;
   npth_rwlock_t rwlock;
-  npth_rwlockattr_t attr;
-  int attr_allocated;
-
-  if (user_attr != NULL)
-    {
-      attr = *user_attr;
-      attr_allocated = 0;
-    }
-  else
-    {
-      err = npth_rwlockattr_init (&attr);
-      if (err)
-	return err;
-    }
 
   /* We can not check *rwlock_r here, as it may contain random data.  */
   rwlock = malloc (sizeof (*rwlock));
@@ -1392,7 +1371,7 @@ npth_rwlock_init (npth_rwlock_t *rwlock_r,
       goto err_out;
     }
 
-  rwlock->prefer_writer = (attr->kind == NPTH_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP);
+  rwlock->prefer_writer = (user_attr && (*user_attr)->kind == NPTH_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP);
 
   err = npth_cond_init (&rwlock->reader_wait, NULL);
   if (err)
@@ -1417,12 +1396,11 @@ npth_rwlock_init (npth_rwlock_t *rwlock_r,
   *rwlock_r = rwlock;
 
  err_out:
-  if (attr_allocated)
-    npth_rwlockattr_destroy (&attr);
   return err;
 }
 
 
+#if 0 /* Not used.  */
 /* Must be called with global lock held.  */
 static int
 rwlock_init_check (npth_rwlock_t *rwlock)
@@ -1461,7 +1439,7 @@ rwlock_init_check (npth_rwlock_t *rwlock)
 
   return err;
 }
-
+#endif
 
 int
 npth_rwlock_destroy (npth_rwlock_t *rwlock)
@@ -1485,7 +1463,7 @@ npth_rwlock_destroy (npth_rwlock_t *rwlock)
     /* FIXME: Log this.  */
     ;
 
-  free (rwlock);
+  free (*rwlock);
 
   *rwlock = NULL;
   return 0;
@@ -1810,9 +1788,7 @@ npth_eselect(int nfd, fd_set *rfds, fd_set *wfds, fd_set *efds,
   int nr_obj = 0;
   /* Number of extra events.  */
   int nr_events = 0;
-  HANDLE sock_event = INVALID_HANDLE_VALUE;
-  /* This will be (nr_obj - 1) == nr_events.  */
-  int sock_event_idx = -1;
+  HANDLE sock_event = NULL;
   int res;
   DWORD ret;
   SOCKET fd;
@@ -1882,13 +1858,12 @@ npth_eselect(int nfd, fd_set *rfds, fd_set *wfds, fd_set *efds,
      return an error.  */
 
   sock_event = WSACreateEvent ();
-  if (sock_event == INVALID_HANDLE_VALUE)
+  if (sock_event == NULL)
     {
       err = EINVAL;
       return -1;
     }
 
-  sock_event_idx = nr_obj;
   obj[nr_obj] = sock_event;
   nr_obj++;
 
@@ -2018,7 +1993,7 @@ npth_eselect(int nfd, fd_set *rfds, fd_set *wfds, fd_set *efds,
 
   /* Cleanup.  */
  err_out:
-  if (sock_event != INVALID_HANDLE_VALUE)
+  if (sock_event != NULL)
     {
       for (i = 0; i < nr_fdobj; i++)
 	{
@@ -2033,3 +2008,7 @@ npth_eselect(int nfd, fd_set *rfds, fd_set *wfds, fd_set *efds,
   errno = err;
   return -1;
 }
+
+/* Include that simple function from the Unix version. */
+#include "../src/getversion.c"
+/* end of file */
